from pathlib import Path

import yaml

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.utils.validation import load_and_validate_platform_config


def get_environment_pipeline_names():
    if not Path(PLATFORM_CONFIG_FILE).exists():
        return {}

    config = load_and_validate_platform_config(disable_aws_validation=True, disable_file_check=True)
    return config.get("environment_pipelines", {}).keys()


def is_terraform_project() -> bool:
    config = yaml.safe_load(Path(PLATFORM_CONFIG_FILE).read_text())
    return not config.get("legacy_project", False)



import os
import re
import subprocess
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version
from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Union

import click
import requests

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_FILE
from dbt_platform_helper.exceptions import IncompatibleMajorVersion
from dbt_platform_helper.exceptions import IncompatibleMinorVersion
from dbt_platform_helper.exceptions import ValidationException
from dbt_platform_helper.utils.validation import load_and_validate_platform_config

VersionTuple = Optional[Tuple[int, int, int]]


class Versions:
    def __init__(self, local_version: VersionTuple = None, latest_release: VersionTuple = None):
        self.local_version = local_version
        self.latest_release = latest_release


class PlatformHelperVersions:
    def __init__(
        self,
        local_version: VersionTuple = None,
        latest_release: VersionTuple = None,
        platform_helper_file_version: VersionTuple = None,
        platform_config_default: VersionTuple = None,
        pipeline_overrides: dict[str, str] = None,
    ):
        self.local_version = local_version
        self.latest_release = latest_release
        self.platform_helper_file_version = platform_helper_file_version
        self.platform_config_default = platform_config_default
        self.pipeline_overrides = pipeline_overrides if pipeline_overrides else {}


def string_version(input_version: VersionTuple) -> str:
    if input_version is None:
        return "unknown"
    major, minor, patch = input_version
    return ".".join([str(s) for s in [major, minor, patch]])


def parse_version(input_version: Union[str, None]) -> VersionTuple:
    if input_version is None:
        return None

    version_plain = input_version.replace("v", "")
    version_segments = re.split(r"[.\-]", version_plain)

    if len(version_segments) != 3:
        return None

    output_version = [0, 0, 0]
    for index, segment in enumerate(version_segments):
        try:
            output_version[index] = int(segment)
        except ValueError:
            output_version[index] = -1
    return output_version[0], output_version[1], output_version[2]


def get_copilot_versions() -> Versions:
    copilot_version = None

    try:
        response = subprocess.run("copilot --version", capture_output=True, shell=True)
        [copilot_version] = re.findall(r"[0-9.]+", response.stdout.decode("utf8"))
    except ValueError:
        pass

    return Versions(parse_version(copilot_version), get_github_released_version("aws/copilot-cli"))


def get_aws_versions() -> Versions:
    aws_version = None
    try:
        response = subprocess.run("aws --version", capture_output=True, shell=True)
        matched = re.match(r"aws-cli/([0-9.]+)", response.stdout.decode("utf8"))
        aws_version = parse_version(matched.group(1))
    except ValueError:
        pass

    return Versions(aws_version, get_github_released_version("aws/aws-cli", True))


def get_github_released_version(repository: str, tags: bool = False) -> Tuple[int, int, int]:
    if tags:
        tags_list = requests.get(f"https://api.github.com/repos/{repository}/tags").json()
        versions = [parse_version(v["name"]) for v in tags_list]
        versions.sort(reverse=True)
        return versions[0]

    package_info = requests.get(f"https://api.github.com/repos/{repository}/releases/latest").json()
    return parse_version(package_info["tag_name"])


def get_platform_helper_versions(include_project_versions=True) -> PlatformHelperVersions:
    try:
        locally_installed_version = parse_version(version("dbt-platform-helper"))
    except PackageNotFoundError:
        locally_installed_version = None

    package_info = requests.get("https://pypi.org/pypi/dbt-platform-helper/json").json()
    released_versions = package_info["releases"].keys()
    parsed_released_versions = [parse_version(v) for v in released_versions]
    parsed_released_versions.sort(reverse=True)
    latest_release = parsed_released_versions[0]

    platform_config_default, pipeline_overrides, version_from_file = None, {}, None

    if include_project_versions:
        platform_config = load_and_validate_platform_config(disable_aws_validation=True)
        platform_config_default = parse_version(
            platform_config.get("default_versions", {}).get("platform-helper")
        )

        pipeline_overrides = {
            name: pipeline.get("versions", {}).get("platform-helper")
            for name, pipeline in platform_config.get("environment_pipelines", {}).items()
            if pipeline.get("versions", {}).get("platform-helper")
        }

        deprecated_version_file = Path(PLATFORM_HELPER_VERSION_FILE)
        version_from_file = (
            parse_version(deprecated_version_file.read_text())
            if deprecated_version_file.exists()
            else None
        )

    out = PlatformHelperVersions(
        local_version=locally_installed_version,
        latest_release=latest_release,
        platform_helper_file_version=version_from_file,
        platform_config_default=platform_config_default,
        pipeline_overrides=pipeline_overrides,
    )

    if include_project_versions:
        _process_version_file_warnings(out)

    return out


def _process_version_file_warnings(versions: PlatformHelperVersions):
    messages = []
    missing_default_version_message = f"Create a section in the root of '{PLATFORM_CONFIG_FILE}':\n\ndefault_versions:\n  platform-helper: "
    deprecation_message = f"Please delete '{PLATFORM_HELPER_VERSION_FILE}' as it is now deprecated."

    if versions.platform_config_default and versions.platform_helper_file_version:
        messages.append(deprecation_message)

    if versions.platform_config_default and not versions.platform_helper_file_version:
        return

    if not versions.platform_config_default and versions.platform_helper_file_version:
        messages.append(deprecation_message)
        messages.append(
            f"{missing_default_version_message}{string_version(versions.platform_helper_file_version)}\n"
        )

    if not versions.platform_config_default and not versions.platform_helper_file_version:
        message = f"Cannot get dbt-platform-helper version from '{PLATFORM_CONFIG_FILE}'.\n"
        message += f"{missing_default_version_message}{string_version(versions.local_version)}\n"
        click.secho(message, fg="red")

    if messages:
        click.secho("\n".join(messages), fg="yellow")


def validate_version_compatibility(
    app_version: Tuple[int, int, int], check_version: Tuple[int, int, int]
):
    app_major, app_minor, app_patch = app_version
    check_major, check_minor, check_patch = check_version
    app_version_as_string = string_version(app_version)
    check_version_as_string = string_version(check_version)

    if (app_major == 0 and check_major == 0) and (
        app_minor != check_minor or app_patch != check_patch
    ):
        raise IncompatibleMajorVersion(app_version_as_string, check_version_as_string)

    if app_major != check_major:
        raise IncompatibleMajorVersion(app_version_as_string, check_version_as_string)

    if app_minor != check_minor:
        raise IncompatibleMinorVersion(app_version_as_string, check_version_as_string)


def check_version_on_file_compatibility(
    app_version: Tuple[int, int, int], file_version: Tuple[int, int, int]
):
    app_major, app_minor, app_patch = app_version
    file_major, file_minor, file_patch = file_version

    return app_major == file_major and app_minor == file_minor and app_patch == file_patch


def get_template_generated_with_version(template_file_path: str) -> Tuple[int, int, int]:
    try:
        template_contents = Path(template_file_path).read_text()
        template_version = re.match(
            r"# Generated by platform-helper ([v.\-0-9]+)", template_contents
        ).group(1)
        return parse_version(template_version)
    except (IndexError, AttributeError):
        raise ValidationException(f"Template {template_file_path} has no version information")


def validate_template_version(app_version: Tuple[int, int, int], template_file_path: str):
    validate_version_compatibility(
        app_version,
        get_template_generated_with_version(template_file_path),
    )


def check_platform_helper_version_needs_update():
    if not running_as_installed_package() or "PLATFORM_TOOLS_SKIP_VERSION_CHECK" in os.environ:
        return

    versions = get_platform_helper_versions(include_project_versions=False)
    local_version = versions.local_version
    latest_release = versions.latest_release
    message = (
        f"You are running platform-helper v{string_version(local_version)}, upgrade to "
        f"v{string_version(latest_release)} by running run `pip install "
        "--upgrade dbt-platform-helper`."
    )
    try:
        validate_version_compatibility(local_version, latest_release)
    except IncompatibleMajorVersion:
        click.secho(message, fg="red")
    except IncompatibleMinorVersion:
        click.secho(message, fg="yellow")


def check_platform_helper_version_mismatch():
    if not running_as_installed_package():
        return

    versions = get_platform_helper_versions()
    local_version = versions.local_version
    platform_helper_file_version = parse_version(
        get_required_platform_helper_version(versions=versions)
    )

    if not check_version_on_file_compatibility(local_version, platform_helper_file_version):
        message = (
            f"WARNING: You are running platform-helper v{string_version(local_version)} against "
            f"v{string_version(platform_helper_file_version)} specified by {PLATFORM_HELPER_VERSION_FILE}."
        )
        click.secho(message, fg="red")


def running_as_installed_package():
    return "site-packages" in __file__


def get_required_platform_helper_version(
    pipeline: str = None, versions: PlatformHelperVersions = None
) -> str:
    if not versions:
        versions = get_platform_helper_versions()
    pipeline_version = versions.pipeline_overrides.get(pipeline)
    version_precedence = [
        pipeline_version,
        versions.platform_config_default,
        versions.platform_helper_file_version,
    ]
    non_null_version_precedence = [
        string_version(v) if isinstance(v, tuple) else v for v in version_precedence if v
    ]

    out = non_null_version_precedence[0] if non_null_version_precedence else None

    if not out:
        raise SystemExit(1)

    return out



from dbt_platform_helper.exceptions import ValidationException


class ARN:
    def __init__(self, arn, *args, **kwargs):
        # store the original ARN
        self.__source = arn

        arn_parts = arn.split(":", 7)

        if len(arn_parts) != 7:
            raise ValidationException(f"Invalid ARN: {arn}")

        # parse and store ARN parts
        # arn:partition:service:region:account-id:resource-type:resource-id
        (
            _,
            partition,
            service,
            region,
            account,
            project,
            build_id,
        ) = arn_parts

        self.__partition = partition
        self.__service = service
        self.__region = region
        self.__account_id = account
        self.__project = project
        self.__build_id = build_id

    @property
    def source(self):
        return self.__source

    @property
    def partition(self):
        return self.__partition

    @property
    def service(self):
        return self.__service

    @property
    def region(self):
        return self.__region

    @property
    def account_id(self):
        return self.__account_id

    @property
    def project(self):
        return self.__project

    @property
    def build_id(self):
        return self.__build_id



import glob
from subprocess import run

import click


def get_lint_result(path: str, ignore_path: str = None, ignore_checks: str = None):
    command = ["cfn-lint", path]
    if ignore_path:
        command.extend(["--ignore-templates", ignore_path])
    if ignore_checks:
        command.extend(["--ignore-checks", ignore_checks])

    click.secho(f"\n>>> Running lint check", fg="yellow")
    click.secho(f"""    {" ".join(command)}\n""", fg="yellow")

    return run(command, capture_output=True)


def get_check_security_result(path: str, ignore_path: str = None):
    matching_files = glob.glob(path)
    command = ["checkov", "--quiet", "--framework", "cloudformation"]

    for file in matching_files:
        command.extend(["--file", file])

    if ignore_path:
        for ignored_file in glob.glob(ignore_path):
            command.extend(["--skip-path", ignore_path])

    click.secho(f"\n>>> Running security check", fg="yellow")
    click.secho(f"""    {" ".join(command)}\n""", fg="yellow")

    return run(command, capture_output=True)



#!/usr/bin/env python

import copy
import json
from os import listdir
from os.path import isfile
from pathlib import Path
from pathlib import PosixPath

import click
import yaml

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.utils.application import get_application_name
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.files import generate_override_files
from dbt_platform_helper.utils.files import mkfile
from dbt_platform_helper.utils.platform_config import is_terraform_project
from dbt_platform_helper.utils.template import camel_case
from dbt_platform_helper.utils.template import setup_templates
from dbt_platform_helper.utils.validation import config_file_check
from dbt_platform_helper.utils.validation import validate_addons
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)

PACKAGE_DIR = Path(__file__).resolve().parent.parent

WAF_ACL_ARN_KEY = "waf-acl-arn"

SERVICE_TYPES = [
    "Load Balanced Web Service",
    "Backend Service",
    "Request-Driven Web Service",
    "Static Site",
    "Worker Service",
]


def list_copilot_local_environments():
    return [
        path.parent.parts[-1] for path in Path("./copilot/environments/").glob("*/manifest.yml")
    ]


def is_service(path: PosixPath) -> bool:
    with open(path) as manifest_file:
        data = yaml.safe_load(manifest_file)
        if not data or not data.get("type"):
            click.echo(
                click.style(f"No type defined in manifest file {str(path)}; exiting", fg="red")
            )
            exit(1)

        return data.get("type") in SERVICE_TYPES


def list_copilot_local_services():
    return [
        path.parent.parts[-1]
        for path in Path("./copilot/").glob("*/manifest.yml")
        if is_service(path)
    ]


@click.group(chain=True, cls=ClickDocOptGroup)
def copilot():
    check_platform_helper_version_needs_update()


def _validate_and_normalise_extensions_config(config_file, key_in_config_file=None):
    """Load a config file, validate it against the extensions schemas and return
    the normalised config dict."""

    def _lookup_plan(addon_type, env_conf):
        plan = env_conf.pop("plan", None)
        conf = addon_plans[addon_type][plan] if plan else {}

        # Make a copy of the addon plan config so subsequent
        # calls do not override the root object
        conf = conf.copy()

        conf.update(env_conf)

        return conf

    def _normalise_keys(source: dict):
        return {k.replace("-", "_"): v for k, v in source.items()}

    with open(PACKAGE_DIR / "addon-plans.yml", "r") as fd:
        addon_plans = yaml.safe_load(fd)

    # load and validate config
    with open(config_file, "r") as fd:
        config = yaml.safe_load(fd)

    if config and key_in_config_file:
        config = config[key_in_config_file]

    # empty file
    if not config:
        return {}

    errors = validate_addons(config)

    if errors:
        click.echo(click.style(f"Errors found in {config_file}:", fg="red"))
        for addon, error in errors.items():
            click.echo(click.style(f"Addon '{addon}': {error}", fg="red"))
        exit(1)

    env_names = list_copilot_local_environments()
    svc_names = list_copilot_local_services()

    if not env_names:
        click.echo(
            click.style(f"No environments found in ./copilot/environments; exiting", fg="red")
        )
        exit(1)

    if not svc_names:
        click.echo(click.style(f"No services found in ./copilot/; exiting", fg="red"))
        exit(1)

    normalised_config = {}
    config_has_errors = False
    for addon_name, addon_config in config.items():
        addon_type = addon_config["type"]
        normalised_config[addon_name] = copy.deepcopy(addon_config)

        if "services" in normalised_config[addon_name]:
            if normalised_config[addon_name]["services"] == "__all__":
                normalised_config[addon_name]["services"] = svc_names

            if not set(normalised_config[addon_name]["services"]).issubset(set(svc_names)):
                click.echo(
                    click.style(
                        f"Services listed in {addon_name}.services do not exist in ./copilot/",
                        fg="red",
                    ),
                )
                config_has_errors = True

        environments = normalised_config[addon_name].pop("environments", {})
        default = environments.pop("*", environments.pop("default", {}))

        initial = _lookup_plan(addon_type, default)

        missing_envs = set(environments.keys()) - set(env_names)
        if missing_envs:
            click.echo(
                click.style(
                    f"Environment keys listed in {addon_name} do not match those defined in ./copilot/environments.",
                    fg="red",
                )
            ),
            click.echo(
                click.style(
                    f"  Missing environments: {', '.join(sorted(missing_envs))}",
                    fg="white",
                ),
            )
            config_has_errors = True

        if config_has_errors:
            continue

        normalised_environments = {}

        for env in env_names:
            normalised_environments[env] = _normalise_keys(initial)

        for env_name, env_config in environments.items():
            if env_config is None:
                env_config = {}
            normalised_environments[env_name].update(
                _lookup_plan(addon_type, _normalise_keys(env_config))
            )

        normalised_config[addon_name]["environments"] = normalised_environments

    if config_has_errors:
        exit(1)

    return normalised_config


def get_log_destination_arn():
    """Get destination arns stored in param store in projects aws account."""
    session = get_aws_session_or_abort()
    client = session.client("ssm", region_name="eu-west-2")
    response = client.get_parameters(Names=["/copilot/tools/central_log_groups"])

    if not response["Parameters"]:
        click.echo(
            click.style(
                "No aws central log group defined in Parameter Store at location /copilot/tools/central_log_groups; exiting",
                fg="red",
            )
        )
        exit(1)

    destination_arns = json.loads(response["Parameters"][0]["Value"])
    return destination_arns


def _generate_svc_overrides(base_path, templates, name):
    click.echo(f"\n>>> Generating service overrides for {name}\n")
    overrides_path = base_path.joinpath(f"copilot/{name}/overrides")
    overrides_path.mkdir(parents=True, exist_ok=True)
    overrides_file = overrides_path.joinpath("cfn.patches.yml")
    overrides_file.write_text(templates.get_template("svc/overrides/cfn.patches.yml").render())


def _get_s3_kms_alias_arns(session, application_name, config):
    application = load_application(application_name, session)
    arns = {}

    for environment_name in application.environments:
        if environment_name not in config:
            continue

        if "bucket_name" not in config[environment_name]:
            continue

        bucket_name = config[environment_name]["bucket_name"]
        kms_client = application.environments[environment_name].session.client("kms")
        alias_name = f"alias/{application_name}-{environment_name}-{bucket_name}-key"

        try:
            response = kms_client.describe_key(KeyId=alias_name)
        except kms_client.exceptions.NotFoundException:
            pass
        else:
            arns[environment_name] = response["KeyMetadata"]["Arn"]

    return arns


@copilot.command()
def make_addons():
    """Generate addons CloudFormation for each environment."""
    output_dir = Path(".").absolute()
    config_file_check()
    is_terraform = is_terraform_project()

    templates = setup_templates()
    config = _get_config()
    session = get_aws_session_or_abort()

    application_name = get_application_name()

    with open(PACKAGE_DIR / "addons-template-map.yml") as fd:
        addon_template_map = yaml.safe_load(fd)

    if is_terraform:
        click.echo("\n>>> Generating Terraform compatible addons CloudFormation\n")
    else:
        click.echo("\n>>> Generating addons CloudFormation\n")

    env_path = Path(f"copilot/environments/")
    env_addons_path = env_path / "addons"
    env_overrides_path = env_path / "overrides"

    (output_dir / env_addons_path).mkdir(parents=True, exist_ok=True)

    _cleanup_old_files(config, output_dir, env_addons_path, env_overrides_path)
    _generate_env_overrides(output_dir, is_terraform)
    custom_resources = _get_custom_resources()

    svc_names = list_copilot_local_services()
    base_path = Path(".")
    for svc_name in svc_names:
        _generate_svc_overrides(base_path, templates, svc_name)

    services = []
    has_addons_parameters = False
    has_postgres_addon = False
    for addon_name, addon_config in config.items():
        print(f">>>>>>>>> {addon_name}")
        addon_type = addon_config.pop("type")
        environments = addon_config.pop("environments")
        if addon_template_map[addon_type].get("requires_addons_parameters", False):
            has_addons_parameters = True
        if addon_type in ["aurora-postgres", "postgres"]:
            has_postgres_addon = True

        for environment_name, environment_config in environments.items():
            if not environment_config.get("deletion_policy"):
                environments[environment_name]["deletion_policy"] = addon_config.get(
                    "deletion_policy", "Delete"
                )

        environment_addon_config = {
            "addon_type": addon_type,
            "custom_resources": custom_resources,
            "environments": environments,
            "name": addon_config.get("name", None) or addon_name,
            "prefix": camel_case(addon_name),
            "secret_name": addon_name.upper().replace("-", "_"),
            **addon_config,
        }

        services.append(environment_addon_config)

        service_addon_config = {
            "application_name": application_name,
            "name": addon_config.get("name", None) or addon_name,
            "prefix": camel_case(addon_name),
            "environments": environments,
            **addon_config,
        }

        log_destination_arns = get_log_destination_arn()

        if addon_type in ["s3", "s3-policy"]:
            if config[addon_name].get("serve_static_content"):
                continue
            if is_terraform:
                s3_kms_arns = _get_s3_kms_alias_arns(session, application_name, environments)
                for environment_name in environments:
                    environments[environment_name]["kms_key_arn"] = s3_kms_arns.get(
                        environment_name, "kms-key-not-found"
                    )
            else:
                service_addon_config["kms_key_reference"] = service_addon_config["prefix"].rsplit(
                    "BucketAccess", 1
                )[0]

        if not is_terraform:
            _generate_env_addons(
                addon_name,
                addon_template_map,
                config.items(),
                env_addons_path,
                environment_addon_config,
                output_dir,
                templates,
                log_destination_arns,
            )
        _generate_service_addons(
            addon_config,
            addon_name,
            addon_template_map,
            addon_type,
            output_dir,
            service_addon_config,
            templates,
            log_destination_arns,
            is_terraform=is_terraform,
        )

        if addon_type in ["aurora-postgres", "postgres"] and not is_terraform:
            click.secho(
                "\nNote: The key DATABASE_CREDENTIALS may need to be changed to match your Django settings configuration.",
                fg="yellow",
            )

    if has_addons_parameters and not is_terraform:
        template = templates.get_template("addons/env/addons.parameters.yml")
        contents = template.render({"has_postgres_addon": has_postgres_addon})
        click.echo(
            mkfile(output_dir, env_addons_path / "addons.parameters.yml", contents, overwrite=True)
        )

    click.echo(templates.get_template("addon-instructions.txt").render(services=services))


def _get_config():
    config = _validate_and_normalise_extensions_config(PACKAGE_DIR / "default-extensions.yml")
    project_config = _validate_and_normalise_extensions_config(PLATFORM_CONFIG_FILE, "extensions")
    config.update(project_config)
    return config


def _generate_env_overrides(output_dir, is_terraform):
    path = "templates/env/terraform-overrides" if is_terraform else "templates/env/overrides"
    click.echo("\n>>> Generating Environment overrides\n")
    overrides_path = output_dir.joinpath(f"copilot/environments/overrides")
    overrides_path.mkdir(parents=True, exist_ok=True)
    template_overrides_path = Path(__file__).parent.parent.joinpath(path)
    generate_override_files(Path("."), template_overrides_path, overrides_path)


def _generate_env_addons(
    addon_name,
    addon_template_map,
    addons,
    env_addons_path,
    environment_addon_config,
    output_dir,
    templates,
    log_destination_arns,
):
    # generate env addons
    addon_type = environment_addon_config["addon_type"]
    for addon in addon_template_map[addon_type].get("env", []):
        template = templates.get_template(addon["template"])
        contents = template.render(
            {
                "addon_config": environment_addon_config,
                "addons": addons,
                "log_destination": log_destination_arns,
            }
        )

        filename = addon.get("filename", f"{addon_name}.yml")

        click.echo(mkfile(output_dir, env_addons_path / filename, contents, overwrite=True))


def _generate_service_addons(
    addon_config,
    addon_name,
    addon_template_map,
    addon_type,
    output_dir,
    service_addon_config,
    templates,
    log_destination_arns,
    is_terraform,
):
    # generate svc addons
    for addon in addon_template_map[addon_type].get("svc", []):
        template = templates.get_template(addon["template"])

        for svc in addon_config.get("services", []):
            service_path = Path(f"copilot/{svc}/addons/")

            contents = template.render(
                {
                    "addon_config": service_addon_config,
                    "log_destination": log_destination_arns,
                    "is_terraform": is_terraform,
                }
            )

            filename = addon.get("filename", f"{addon_name}.yml")

            (output_dir / service_path).mkdir(parents=True, exist_ok=True)
            click.echo(mkfile(output_dir, service_path / filename, contents, overwrite=True))


def _get_custom_resources():
    custom_resources = {}
    custom_resource_path = (Path(__file__).parent / "../custom_resources/").resolve()
    for file in listdir(custom_resource_path):
        file_path = custom_resource_path / file
        if isfile(file_path) and file_path.name.endswith(".py") and file_path.name != "__init__.py":
            custom_resource_contents = file_path.read_text()

            def file_with_formatting_options(padding=0):
                lines = [
                    (" " * padding) + line if line.strip() else line.strip()
                    for line in custom_resource_contents.splitlines(True)
                ]
                return "".join(lines)

            custom_resources[file_path.name.rstrip(".py")] = file_with_formatting_options
    return custom_resources


def _cleanup_old_files(config, output_dir, env_addons_path, env_overrides_path):
    def _rmdir(path):
        if not path.exists():
            return
        for f in path.iterdir():
            if f.is_file():
                f.unlink()
            if f.is_dir():
                _rmdir(f)
                f.rmdir()

    _rmdir(output_dir / env_addons_path)
    _rmdir(output_dir / env_overrides_path)

    all_services = set()
    for services in [v["services"] for v in config.values() if "services" in v]:
        all_services.update(services)

    for service in all_services:
        svc_addons_dir = Path(output_dir, "copilot", service, "addons")
        if not svc_addons_dir.exists():
            continue
        for f in svc_addons_dir.iterdir():
            if f.is_file():
                f.unlink()



import json
import os
import urllib.parse
from configparser import ConfigParser
from pathlib import Path
from typing import Tuple

import boto3
import botocore
import click
import yaml
from boto3 import Session

from dbt_platform_helper.exceptions import ValidationException

SSM_BASE_PATH = "/copilot/{app}/{env}/secrets/"
SSM_PATH = "/copilot/{app}/{env}/secrets/{name}"
AWS_SESSION_CACHE = {}


def get_aws_session_or_abort(aws_profile: str = None) -> boto3.session.Session:
    aws_profile = aws_profile if aws_profile else os.getenv("AWS_PROFILE")
    if aws_profile in AWS_SESSION_CACHE:
        return AWS_SESSION_CACHE[aws_profile]

    # Check that the aws profile exists and is set.
    click.secho(f"""Checking AWS connection for profile "{aws_profile}"...""", fg="cyan")

    try:
        session = boto3.session.Session(profile_name=aws_profile)
    except botocore.exceptions.ProfileNotFound:
        click.secho(f"""AWS profile "{aws_profile}" is not configured.""", fg="red")
        exit(1)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "ExpiredToken":
            click.secho(
                f"Credentials are NOT valid.  \nPlease login with: aws sso login --profile {aws_profile}",
                fg="red",
            )
            exit(1)

    sts = session.client("sts")
    try:
        account_id, user_id = get_account_details(sts)
        click.secho("Credentials are valid.", fg="green")
    except (
        botocore.exceptions.UnauthorizedSSOTokenError,
        botocore.exceptions.TokenRetrievalError,
        botocore.exceptions.SSOTokenLoadError,
    ):
        click.secho(
            "The SSO session associated with this profile has expired or is otherwise invalid."
            "To refresh this SSO session run `aws sso login` with the corresponding profile",
            fg="red",
        )
        exit(1)

    alias_client = session.client("iam")
    account_name = alias_client.list_account_aliases()["AccountAliases"]
    if account_name:
        click.echo(
            click.style("Logged in with AWS account: ", fg="yellow")
            + click.style(f"{account_name[0]}/{account_id}", fg="white", bold=True),
        )
    else:
        click.echo(
            click.style("Logged in with AWS account id: ", fg="yellow")
            + click.style(f"{account_id}", fg="white", bold=True),
        )
    click.echo(
        click.style("User: ", fg="yellow")
        + click.style(f"{user_id.split(':')[-1]}\n", fg="white", bold=True),
    )

    AWS_SESSION_CACHE[aws_profile] = session

    return session


class NoProfileForAccountIdError(Exception):
    def __init__(self, account_id):
        super().__init__(f"No profile found for account {account_id}")


def get_profile_name_from_account_id(account_id: str):
    aws_config = ConfigParser()
    aws_config.read(Path.home().joinpath(".aws/config"))
    for section in aws_config.sections():
        found_account_id = aws_config[section].get("sso_account_id", None)
        if account_id == found_account_id:
            return section.removeprefix("profile ")

    raise NoProfileForAccountIdError(account_id)


def get_ssm_secret_names(app, env):
    session = get_aws_session_or_abort()
    client = session.client("ssm")

    path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secret_names = []

    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secret_names.append(secret["Name"])

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secret_names)


def get_ssm_secrets(app, env, session=None, path=None):
    """Return secrets from AWS Parameter Store as a list of tuples with the
    secret name and secret value."""

    if not session:
        session = get_aws_session_or_abort()
    client = session.client("ssm")

    if not path:
        path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secrets = []

    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secrets.append((secret["Name"], secret["Value"]))

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secrets)


def set_ssm_param(
    app, env, param_name, param_value, overwrite, exists, description="Copied from Cloud Foundry."
):
    session = get_aws_session_or_abort()
    client = session.client("ssm")

    parameter_args = dict(
        Name=param_name,
        Description=description,
        Value=param_value,
        Type="SecureString",
        Overwrite=overwrite,
        Tags=[
            {"Key": "copilot-application", "Value": app},
            {"Key": "copilot-environment", "Value": env},
        ],
        Tier="Intelligent-Tiering",
    )

    if overwrite and not exists:
        raise ValidationException(
            """Arguments "overwrite" is set to True, but "exists" is set to False."""
        )

    if overwrite and exists:
        # Tags can't be updated when overwriting
        del parameter_args["Tags"]

    client.put_parameter(**parameter_args)


def check_response(response):
    if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
        click.secho(
            f"Unknown response error from AWS.\nStatus Code: {response['ResponseMetadata']['HTTPStatusCode']}",
            fg="red",
        )
        exit()


def get_codestar_connection_arn(app_name):
    session = get_aws_session_or_abort()
    response = session.client("codestar-connections").list_connections()

    for connection in response["Connections"]:
        if connection["ConnectionName"] == app_name:
            return connection["ConnectionArn"]


def get_account_details(sts_client=None):
    if not sts_client:
        sts_client = get_aws_session_or_abort().client("sts")
    response = sts_client.get_caller_identity()

    return response["Account"], response["UserId"]


def get_public_repository_arn(repository_uri):
    session = get_aws_session_or_abort()
    response = session.client("ecr-public", region_name="us-east-1").describe_repositories()
    repository = [
        repo for repo in response["repositories"] if repo["repositoryUri"] == repository_uri
    ]

    return repository[0]["repositoryArn"] if repository else None


def get_load_balancer_domain_and_configuration(
    project_session: Session, app: str, env: str, svc: str
) -> Tuple[str, dict]:
    response = get_load_balancer_configuration(project_session, app, env, svc)

    # Find the domain name
    with open(f"./copilot/{svc}/manifest.yml", "r") as fd:
        conf = yaml.safe_load(fd)
        if "environments" in conf:
            if env in conf["environments"]:
                for domain in conf["environments"].items():
                    if domain[0] == env:
                        if (
                            domain[1] is None
                            or domain[1]["http"] is None
                            or domain[1]["http"]["alias"] is None
                        ):
                            click.secho(
                                f"No domains found, please check the ./copilot/{svc}/manifest.yml file",
                                fg="red",
                            )
                            exit()
                        domain_name = domain[1]["http"]["alias"]
            else:
                click.secho(
                    f"Environment {env} not found, please check the ./copilot/{svc}/manifest.yml file",
                    fg="red",
                )
                exit()

    return domain_name, response["LoadBalancers"][0]


def get_load_balancer_configuration(
    project_session: Session, app: str, env: str, svc: str
) -> list[Session]:
    proj_client = project_session.client("ecs")

    response = proj_client.list_clusters()
    check_response(response)
    no_items = True
    for cluster_arn in response["clusterArns"]:
        cluster_name = cluster_arn.split("/")[1]
        if cluster_name.startswith(f"{app}-{env}-Cluster"):
            no_items = False
            break

    if no_items:
        click.echo(
            click.style("There are no clusters for environment ", fg="red")
            + click.style(f"{env} ", fg="white", bold=True)
            + click.style("of application ", fg="red")
            + click.style(f"{app} ", fg="white", bold=True)
            + click.style("in AWS account ", fg="red")
            + click.style(f"{project_session.profile_name}", fg="white", bold=True),
        )
        exit()

    response = proj_client.list_services(cluster=cluster_name)
    check_response(response)
    no_items = True
    for service_arn in response["serviceArns"]:
        fully_qualified_service_name = service_arn.split("/")[2]
        if fully_qualified_service_name.startswith(f"{app}-{env}-{svc}-Service"):
            no_items = False
            break

    if no_items:
        click.echo(
            click.style("There are no services called ", fg="red")
            + click.style(f"{svc} ", fg="white", bold=True)
            + click.style("for environment ", fg="red")
            + click.style(f"{env} ", fg="white", bold=True)
            + click.style("of application ", fg="red")
            + click.style(f"{app} ", fg="white", bold=True)
            + click.style("in AWS account ", fg="red")
            + click.style(f"{project_session.profile_name}", fg="white", bold=True),
        )
        exit()

    elb_client = project_session.client("elbv2")

    elb_arn = elb_client.describe_target_groups(
        TargetGroupArns=[
            proj_client.describe_services(
                cluster=cluster_name,
                services=[
                    fully_qualified_service_name,
                ],
            )["services"][0]["loadBalancers"][0]["targetGroupArn"],
        ],
    )["TargetGroups"][0]["LoadBalancerArns"][0]

    response = elb_client.describe_load_balancers(LoadBalancerArns=[elb_arn])
    check_response(response)
    return response


def update_postgres_parameter_with_master_secret(session, parameter_name, secret_arn):
    ssm_client = session.client("ssm")
    secrets_manager_client = session.client("secretsmanager")
    response = ssm_client.get_parameter(Name=parameter_name, WithDecryption=True)
    parameter_value = response["Parameter"]["Value"]

    parameter_data = json.loads(parameter_value)

    secret_response = secrets_manager_client.get_secret_value(SecretId=secret_arn)
    secret_value = json.loads(secret_response["SecretString"])

    parameter_data["username"] = urllib.parse.quote(secret_value["username"])
    parameter_data["password"] = urllib.parse.quote(secret_value["password"])

    return parameter_data



import yaml


def get_service_name_from_manifest(manifest_path):
    with open(manifest_path) as manifest:
        document = yaml.safe_load(manifest)
        return document["name"]


def get_repository_name_from_manifest(manifest_path):
    with open(manifest_path) as manifest:
        document = yaml.safe_load(manifest)
        image = document["image"]["location"]

        repository_with_tag = image.split("/", 1)[1]
        repository = repository_with_tag.split(":")[0]

        return repository



#!/usr/bin/env python
from pathlib import Path

import click

from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.cloudformation import get_check_security_result
from dbt_platform_helper.utils.cloudformation import get_lint_result

BASE_DIR = Path(__file__).parent.parent.parent


@click.group(invoke_without_command=True, chain=True, cls=ClickDocOptGroup)
@click.option("-d", "--directory", type=str, default="copilot")
@click.pass_context
def check_cloudformation(ctx: click.Context, directory: str) -> None:
    """
    Runs the checks passed in the command arguments.

    If no argument is passed, it will run all the checks.
    """
    ctx.ensure_object(dict)

    if ctx.invoked_subcommand is None:
        click.secho(f"\n>>> Running all checks", fg="yellow")
        for command in ctx.command.commands.values():
            ctx.invoke(command)


@check_cloudformation.command()
@click.option("-d", "--directory", type=str, default="copilot")
@click.pass_context
def lint(ctx: click.Context, directory: str) -> bool:
    """Runs cfn-lint against the generated CloudFormation templates."""
    addons_manifests = f"{directory}/**/addons/*.yml"
    # addons.parameters.yml is not a CloudFormation template file
    ignore_addons_params = f"{directory}/**/addons/addons.parameters.yml"
    # "W2001 Parameter Env not used" is ignored becomes Copilot addons require
    # parameters even if they are not used in the Cloudformation template.
    ignore_checks = "W2001"

    result = get_lint_result(addons_manifests, ignore_addons_params, ignore_checks)
    success = result.returncode == 0

    ctx.obj["lint"] = {
        "success": success,
        "message": result.stdout.decode() if not success else None,
    }

    return success


@check_cloudformation.command()
@click.option("-d", "--directory", type=str, default="copilot")
@click.pass_context
def check_security(ctx: click.Context, directory: str) -> bool:
    addons_manifests = f"{directory}/**/addons/*.yml"
    # # addons.parameters.yml is not a CloudFormation template file
    ignore_addons_params = f"{directory}/**/addons/addons.parameters.yml"

    result = get_check_security_result(addons_manifests, ignore_addons_params)
    success = result.returncode == 0

    ctx.obj["check-security"] = {
        "success": success,
        "message": result.stdout.decode() if not success else None,
    }

    return success


@check_cloudformation.result_callback()
@click.pass_context
def process_result(ctx: click.Context, result, directory) -> None:
    successful = {k: v for k, v in ctx.obj.items() if v["success"]}
    failed = {k: v for k, v in ctx.obj.items() if not v["success"]}
    if successful:
        click.secho("\nThe CloudFormation templates passed the following checks:", fg="green")
        for subcommand_name in successful:
            click.secho(f"  - {subcommand_name}", fg="white")
    if failed:
        click.secho("\nThe CloudFormation templates failed the following checks:", fg="red")
        for subcommand_name in failed:
            message = failed[subcommand_name]["message"]
            click.secho(f"  - {subcommand_name} [{message}]", fg="white")
        exit(1)
    exit(0)



import datetime
from importlib.metadata import version

from jinja2_simple_tags import StandaloneTag


class VersionTag(StandaloneTag):
    tags = {"version_info"}

    def render(self):
        format = "%Y-%m-%d %H:%M:%S"
        time = datetime.datetime.now().strftime(format)
        return f"Generated by platform-helper {version('dbt-platform-helper')} / {time}"


class ExtraHeaderTag(StandaloneTag):
    tags = {"extra_header"}

    def render(self):
        return f"WARNING: This is an autogenerated file, not for manual editing."



import click
from cloudfoundry_client.client import CloudFoundryClient


def get_cloud_foundry_client_or_abort():
    try:
        client = CloudFoundryClient.build_from_cf_config()
        click.secho("Logged in to Cloud Foundry", fg="green")
        return client
    except Exception as ex:
        click.secho("Could not connect to Cloud Foundry: ", fg="red", nl=False)
        click.secho(str(ex))
        click.secho("Please log in with: cf login", fg="yellow")
        exit(1)



import click

from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.platform_config import get_environment_pipeline_names
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)
from dbt_platform_helper.utils.versioning import get_required_platform_helper_version


@click.group(chain=True, cls=ClickDocOptGroup)
def version():
    """Contains subcommands for getting version information about the current
    project."""
    check_platform_helper_version_needs_update()


@version.command(help="Print the version of platform-tools required by the current project")
@click.option(
    "--pipeline",
    required=False,
    type=click.Choice(get_environment_pipeline_names()),
    help="Take into account platform-tools version overrides in the specified pipeline",
)
def get_platform_helper_for_project(pipeline):
    """
    Version precedence is in this order:
        - if the --pipeline option is supplied, the version in 'platform-config.yml' in:
            environment_pipelines:
                <pipeline>:
                    ...
                    versions:
                        platform-helper
        - The version from default_versions/platform-helper in 'platform-config.yml'
        - Fall back on the version in the deprecated '.platform-helper-version' file
    """
    required_version = get_required_platform_helper_version(pipeline)
    click.secho(required_version)



import subprocess
from typing import List

import click

from dbt_platform_helper.commands.conduit import add_stack_delete_policy_to_task_role
from dbt_platform_helper.commands.conduit import addon_client_is_running
from dbt_platform_helper.commands.conduit import connect_to_addon_client_task
from dbt_platform_helper.commands.conduit import get_cluster_arn
from dbt_platform_helper.commands.conduit import normalise_secret_name
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import update_postgres_parameter_with_master_secret
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)


@click.group(chain=True, cls=ClickDocOptGroup)
def database():
    check_platform_helper_version_needs_update()


@database.command(name="copy")
@click.argument("source_db", type=str, required=True)
@click.argument("target_db", type=str, required=True)
def copy(source_db: str, target_db: str):
    """Copy source database to target database."""
    app = None
    source_env = None
    target_env = None

    for tag in get_database_tags(source_db):
        if tag["Key"] == "copilot-application":
            app = tag["Value"]
        if tag["Key"] == "copilot-environment":
            source_env = tag["Value"]
            if app is not None:
                break

    for tag in get_database_tags(target_db):
        if tag["Key"] == "copilot-environment":
            target_env = tag["Value"]
            break

    if not app or not source_env or not target_env:
        click.secho(f"""Required database tags not found.""", fg="red")
        exit(1)

    if target_env == "prod":
        click.secho(f"""The target database cannot be a production database.""", fg="red")
        exit(1)

    if source_db == target_db:
        click.secho(f"""Source and target databases are the same.""", fg="red")
        exit(1)

    if not click.confirm(
        click.style("Copying data from ", fg="yellow")
        + click.style(f"{source_db} ", fg="white", bold=True)
        + click.style(f"in environment {source_env} to ", fg="yellow", bold=True)
        + click.style(f"{target_db} ", fg="white", bold=True)
        + click.style(f"in environment {target_env}\n", fg="yellow", bold=True)
        + click.style("Do you want to continue?", fg="yellow"),
    ):
        exit()

    click.echo(f"""Starting task to copy data from {source_db} to {target_db}""")

    source_db_connection = get_connection_string(app, source_env, source_db)
    target_db_connection = get_connection_string(app, target_env, target_db)

    application = load_application(app)
    cluster_arn = get_cluster_arn(application, source_env)
    task_name = f"database-copy-{app}-{source_env}-{app}-{target_env}"

    if not addon_client_is_running(application, source_env, cluster_arn, task_name):
        subprocess.call(
            f"copilot task run --app {app} --env {source_env} "
            f"--task-group-name {task_name} "
            f"--image public.ecr.aws/uktrade/tunnel:database-copy "
            f"--env-vars SOURCE_DB_CONNECTION='{source_db_connection}',TARGET_DB_CONNECTION='{target_db_connection}' "
            "--platform-os linux "
            "--platform-arch arm64",
            shell=True,
        )
        add_stack_delete_policy_to_task_role(application, source_env, task_name)
    connect_to_addon_client_task(application, source_env, cluster_arn, task_name)


def get_database_tags(db_identifier: str) -> List[dict]:
    session = get_aws_session_or_abort()
    rds = session.client("rds")

    try:
        db_instance = rds.describe_db_instances(DBInstanceIdentifier=db_identifier)["DBInstances"][
            0
        ]

        return db_instance["TagList"]
    except rds.exceptions.DBInstanceNotFoundFault:
        click.secho(
            f"""Database {db_identifier} not found. Check the database identifier.""", fg="red"
        )
        exit(1)


def get_connection_string(app: str, env: str, db_identifier: str) -> str:
    session = get_aws_session_or_abort()
    addon_name = normalise_secret_name(db_identifier.split(f"{app}-{env}-", 1)[1])
    connection_string_parameter = f"/copilot/{app}/{env}/secrets/{addon_name}_READ_ONLY_USER"
    master_secret_name = f"/copilot/{app}/{env}/secrets/{addon_name}_RDS_MASTER_ARN"
    master_secret_arn = session.client("ssm").get_parameter(
        Name=master_secret_name, WithDecryption=True
    )["Parameter"]["Value"]

    conn = update_postgres_parameter_with_master_secret(
        session, connection_string_parameter, master_secret_arn
    )

    return f"postgres://{conn['username']}:{conn['password']}@{conn['host']}:{conn['port']}/{conn['dbname']}"



#!/usr/bin/env python
from os import makedirs
from pathlib import Path
from shutil import rmtree

import click

from dbt_platform_helper.utils.application import get_application_name
from dbt_platform_helper.utils.aws import get_account_details
from dbt_platform_helper.utils.aws import get_codestar_connection_arn
from dbt_platform_helper.utils.aws import get_public_repository_arn
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.files import apply_environment_defaults
from dbt_platform_helper.utils.files import generate_override_files_from_template
from dbt_platform_helper.utils.files import mkfile
from dbt_platform_helper.utils.git import git_remote
from dbt_platform_helper.utils.messages import abort_with_error
from dbt_platform_helper.utils.platform_config import is_terraform_project
from dbt_platform_helper.utils.template import setup_templates
from dbt_platform_helper.utils.validation import load_and_validate_platform_config
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)

CODEBASE_PIPELINES_KEY = "codebase_pipelines"
ENVIRONMENTS_KEY = "environments"


@click.group(chain=True, cls=ClickDocOptGroup)
def pipeline():
    """Pipeline commands."""
    check_platform_helper_version_needs_update()


@pipeline.command()
def generate():
    """Given a platform-config.yml file, generate environment and service
    deployment pipelines."""
    pipeline_config = load_and_validate_platform_config()

    no_codebase_pipelines = CODEBASE_PIPELINES_KEY not in pipeline_config
    no_environment_pipelines = ENVIRONMENTS_KEY not in pipeline_config

    if no_codebase_pipelines and no_environment_pipelines:
        click.secho("No pipelines defined: nothing to do.", err=True, fg="yellow")
        return

    templates = setup_templates()
    app_name = get_application_name()

    git_repo = git_remote()
    if not git_repo:
        abort_with_error("The current directory is not a git repository")

    codestar_connection_arn = get_codestar_connection_arn(app_name)
    if codestar_connection_arn is None:
        abort_with_error(f'There is no CodeStar Connection named "{app_name}" to use')

    base_path = Path(".")
    pipelines_dir = base_path / f"copilot/pipelines"

    _clean_pipeline_config(pipelines_dir)

    if not is_terraform_project() and ENVIRONMENTS_KEY in pipeline_config:
        _generate_copilot_environments_pipeline(
            app_name,
            codestar_connection_arn,
            git_repo,
            apply_environment_defaults(pipeline_config)[ENVIRONMENTS_KEY],
            base_path,
            pipelines_dir,
            templates,
        )

    if CODEBASE_PIPELINES_KEY in pipeline_config:
        account_id, _ = get_account_details()

        for codebase in pipeline_config[CODEBASE_PIPELINES_KEY]:
            _generate_codebase_pipeline(
                account_id,
                app_name,
                codestar_connection_arn,
                git_repo,
                codebase,
                base_path,
                pipelines_dir,
                templates,
            )


def _clean_pipeline_config(pipelines_dir):
    if pipelines_dir.exists():
        click.echo("Deleting copilot/pipelines directory.")
        rmtree(pipelines_dir)


def _generate_codebase_pipeline(
    account_id,
    app_name,
    codestar_connection_arn,
    git_repo,
    codebase,
    base_path,
    pipelines_dir,
    templates,
):
    makedirs(pipelines_dir / codebase["name"] / "overrides", exist_ok=True)
    environments = []
    for pipelines in codebase["pipelines"]:
        environments += pipelines[ENVIRONMENTS_KEY]

    additional_ecr = codebase.get("additional_ecr_repository", None)
    add_public_perms = additional_ecr and additional_ecr.startswith("public.ecr.aws")
    additional_ecr_arn = get_public_repository_arn(additional_ecr) if add_public_perms else None

    template_data = {
        "account_id": account_id,
        "app_name": app_name,
        "deploy_repo": git_repo,
        "codebase": codebase,
        ENVIRONMENTS_KEY: environments,
        "codestar_connection_arn": codestar_connection_arn,
        "codestar_connection_id": codestar_connection_arn.split("/")[-1],
        "additional_ecr_arn": additional_ecr_arn,
    }

    _create_file_from_template(
        base_path,
        f"{codebase['name']}/manifest.yml",
        pipelines_dir,
        template_data,
        templates,
        "codebase/manifest.yml",
    )

    overrides_path = Path(__file__).parent.parent.joinpath("templates/pipelines/codebase/overrides")
    generate_override_files_from_template(
        base_path, overrides_path, pipelines_dir / codebase["name"] / "overrides", template_data
    )


def _generate_copilot_environments_pipeline(
    app_name, codestar_connection_arn, git_repo, configuration, base_path, pipelines_dir, templates
):
    makedirs(pipelines_dir / "environments/overrides", exist_ok=True)

    template_data = {
        "app_name": app_name,
        "git_repo": git_repo,
        "codestar_connection_arn": codestar_connection_arn,
        "pipeline_environments": configuration,
    }

    _create_file_from_template(
        base_path, "environments/buildspec.yml", pipelines_dir, template_data, templates
    )
    _create_file_from_template(
        base_path, "environments/manifest.yml", pipelines_dir, template_data, templates
    )
    _create_file_from_template(
        base_path, "environments/overrides/cfn.patches.yml", pipelines_dir, template_data, templates
    )


def _create_file_from_template(
    base_path, file_name, pipelines_dir, template_data, templates, template_name=None
):
    contents = templates.get_template(
        f"pipelines/{file_name if template_name is None else template_name}"
    ).render(template_data)
    message = mkfile(base_path, pipelines_dir / file_name, contents, overwrite=True)
    click.echo(message)



import json
import re
from pathlib import Path
from typing import Dict

import boto3
import yaml
from boto3 import Session
from yaml.parser import ParserError

from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import get_profile_name_from_account_id
from dbt_platform_helper.utils.aws import get_ssm_secrets
from dbt_platform_helper.utils.messages import abort_with_error


class Environment:
    name: str
    account_id: str
    sessions: Dict[str, boto3.Session]

    def __init__(self, name: str, account_id: str, sessions: Dict[str, boto3.Session]):
        self.name = name
        self.account_id = account_id
        self.sessions = sessions

    @property
    def session(self):
        if self.account_id not in self.sessions:
            self.sessions[self.account_id] = get_aws_session_or_abort(
                get_profile_name_from_account_id(self.account_id),
            )

        return self.sessions[self.account_id]


class Service:
    name: str
    kind: str

    def __init__(self, name: str, kind: str):
        self.name = name
        self.kind = kind


class Application:
    name: str
    environments: Dict[str, Environment]
    services: Dict[str, Service]

    def __init__(self, name: str):
        self.name = name
        self.environments = {}
        self.services = {}

    def __str__(self):
        output = f"Application {self.name} with"

        environments = [f"{env.name}:{env.account_id}" for env in self.environments.values()]

        if environments:
            return f"{output} environments {', '.join(environments)}"

        return f"{output} no environments"

    def __eq__(self, other):
        return str(self) == str(other)


class ApplicationNotFoundError(Exception):
    pass


def load_application(app: str = None, default_session: Session = None) -> Application:
    application = Application(app if app else get_application_name())
    current_session = default_session if default_session else get_aws_session_or_abort()

    ssm_client = current_session.client("ssm")

    try:
        ssm_client.get_parameter(
            Name=f"/copilot/applications/{application.name}",
            WithDecryption=False,
        )
    except ssm_client.exceptions.ParameterNotFound:
        raise ApplicationNotFoundError

    path = f"/copilot/applications/{application.name}/environments"
    secrets = get_ssm_secrets(app, None, current_session, path)

    sts_client = current_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    sessions = {account_id: current_session}

    def is_environment_key(name):
        """
        Match only parameter names that are an environment path with no further
        nesting.

        e.g.
         - /copilot/applications/test/environments/my_env will match.
         - /copilot/applications/test/environments/my_env/addons will not match.
        """
        environment_key_regex = r"^/copilot/applications/{}/environments/[^/]*$".format(
            application.name
        )
        return bool(re.match(environment_key_regex, name))

    environments = {
        env["name"]: Environment(env["name"], env["accountID"], sessions)
        for env in [json.loads(s[1]) for s in secrets if is_environment_key(s[0])]
    }
    application.environments = environments

    response = ssm_client.get_parameters_by_path(
        Path=f"/copilot/applications/{application.name}/components",
        Recursive=False,
        WithDecryption=False,
    )

    application.services = {
        svc["name"]: Service(svc["name"], svc["type"])
        for svc in [json.loads(parameter["Value"]) for parameter in response["Parameters"]]
    }

    return application


def get_application_name():
    app_name = None
    try:
        app_config = yaml.safe_load(Path("copilot/.workspace").read_text())
        app_name = app_config["application"]
    except (FileNotFoundError, ParserError):
        pass

    if app_name is None:
        abort_with_error("Cannot get application name. No copilot/.workspace file found")

    return app_name



from copy import deepcopy
from os import makedirs
from pathlib import Path

import click
import yaml
from jinja2 import Environment
from jinja2 import FileSystemLoader


def to_yaml(value):
    return yaml.dump(value, sort_keys=False)


def mkfile(base_path, file_path, contents, overwrite=False):
    file_path = Path(file_path)
    file = Path(base_path).joinpath(file_path)
    file_exists = file.exists()

    if not file_path.parent.exists():
        makedirs(file_path.parent)

    if file_exists and not overwrite:
        return f"File {file_path} exists; doing nothing"

    action = "overwritten" if file_exists and overwrite else "created"

    file.write_text(contents)

    return f"File {file_path} {action}"


def generate_override_files(base_path, file_path, output_dir):
    def generate_files_for_dir(pattern):
        for file in file_path.glob(pattern):
            if file.is_file():
                contents = file.read_text()
                file_name = str(file).removeprefix(f"{file_path}/")
                click.echo(
                    mkfile(
                        base_path,
                        output_dir / file_name,
                        contents,
                        overwrite=True,
                    )
                )

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")


def generate_override_files_from_template(base_path, overrides_path, output_dir, template_data={}):
    templates = Environment(
        loader=FileSystemLoader(f"{overrides_path}"), keep_trailing_newline=True
    )
    environments = ",".join([env["name"] for env in template_data["environments"]])
    data = {"environments": environments}

    def generate_files_for_dir(pattern):

        for file in overrides_path.glob(pattern):
            if file.is_file():
                file_name = str(file).removeprefix(f"{overrides_path}/")
                contents = templates.get_template(str(file_name)).render(data)
                message = mkfile(base_path, output_dir / file_name, contents, overwrite=True)
                click.echo(message)

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")


def apply_environment_defaults(config):
    if "environments" not in config:
        return config

    enriched_config = deepcopy(config)

    environments = enriched_config["environments"]
    env_defaults = environments.get("*", {})
    without_defaults_entry = {
        name: data if data else {} for name, data in environments.items() if name != "*"
    }

    default_versions = config.get("default_versions", {})

    def combine_env_data(data):
        return {
            **env_defaults,
            **data,
            "versions": {
                **default_versions,
                **env_defaults.get("versions", {}),
                **data.get("versions", {}),
            },
        }

    defaulted_envs = {
        env_name: combine_env_data(env_data)
        for env_name, env_data in without_defaults_entry.items()
    }

    enriched_config["environments"] = defaulted_envs

    return enriched_config



import random
import re
import string
from pathlib import Path
from typing import List
from typing import Union

import boto3
import click
from schema import SchemaError

from dbt_platform_helper.constants import DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.utils.application import Environment
from dbt_platform_helper.utils.application import Service
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.files import apply_environment_defaults
from dbt_platform_helper.utils.files import mkfile
from dbt_platform_helper.utils.platform_config import is_terraform_project
from dbt_platform_helper.utils.template import setup_templates
from dbt_platform_helper.utils.validation import load_and_validate_platform_config
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)

AVAILABLE_TEMPLATES = ["default", "migration", "dmas-migration"]


@click.group(cls=ClickDocOptGroup)
def environment():
    """Commands affecting environments."""
    check_platform_helper_version_needs_update()


@environment.command()
@click.option("--app", type=str, required=True)
@click.option("--env", type=str, required=True)
@click.option("--svc", type=str, required=True, multiple=True, default=["web"])
@click.option(
    "--template",
    type=click.Choice(AVAILABLE_TEMPLATES),
    default="default",
    help="The maintenance page you wish to put up.",
)
@click.option("--vpc", type=str)
def offline(app, env, svc, template, vpc):
    """Take load-balanced web services offline with a maintenance page."""
    application = get_application(app)
    application_environment = get_app_environment(app, env)

    if "*" in svc:
        services = [
            s for s in application.services.values() if s.kind == "Load Balanced Web Service"
        ]
    else:
        all_services = [get_app_service(app, s) for s in list(svc)]
        services = [s for s in all_services if s.kind == "Load Balanced Web Service"]

    if not services:
        click.secho(f"No services deployed yet to {app} environment {env}", fg="red")
        raise click.Abort

    try:
        https_listener = find_https_listener(application_environment.session, app, env)
        current_maintenance_page = get_maintenance_page(
            application_environment.session, https_listener
        )
        remove_current_maintenance_page = False
        if current_maintenance_page:
            remove_current_maintenance_page = click.confirm(
                f"There is currently a '{current_maintenance_page}' maintenance page for the {env} "
                f"environment in {app}.\nWould you like to replace it with a '{template}' "
                f"maintenance page?"
            )
            if not remove_current_maintenance_page:
                raise click.Abort

        if remove_current_maintenance_page or click.confirm(
            f"You are about to enable the '{template}' maintenance page for the {env} "
            f"environment in {app}.\nWould you like to continue?"
        ):
            if current_maintenance_page and remove_current_maintenance_page:
                remove_maintenance_page(application_environment.session, https_listener)

            allowed_ips = get_env_ips(vpc, application_environment)

            add_maintenance_page(
                application_environment.session,
                https_listener,
                app,
                env,
                services,
                allowed_ips,
                template,
            )
            click.secho(
                f"Maintenance page '{template}' added for environment {env} in application {app}",
                fg="green",
            )
        else:
            raise click.Abort

    except LoadBalancerNotFoundError:
        click.secho(
            f"No load balancer found for environment {env} in the application {app}.", fg="red"
        )
        raise click.Abort

    except ListenerNotFoundError:
        click.secho(
            f"No HTTPS listener found for environment {env} in the application {app}.", fg="red"
        )
        raise click.Abort


@environment.command()
@click.option("--app", type=str, required=True)
@click.option("--env", type=str, required=True)
def online(app, env):
    """Remove a maintenance page from an environment."""
    application_environment = get_app_environment(app, env)

    try:
        https_listener = find_https_listener(application_environment.session, app, env)
        current_maintenance_page = get_maintenance_page(
            application_environment.session, https_listener
        )
        if not current_maintenance_page:
            click.secho("There is no current maintenance page to remove", fg="red")
            raise click.Abort

        if not click.confirm(
            f"There is currently a '{current_maintenance_page}' maintenance page, "
            f"would you like to remove it?"
        ):
            raise click.Abort

        remove_maintenance_page(application_environment.session, https_listener)
        click.secho(
            f"Maintenance page removed from environment {env} in application {app}", fg="green"
        )

    except LoadBalancerNotFoundError:
        click.secho(
            f"No load balancer found for environment {env} in the application {app}.", fg="red"
        )
        raise click.Abort

    except ListenerNotFoundError:
        click.secho(
            f"No HTTPS listener found for environment {env} in the application {app}.", fg="red"
        )
        raise click.Abort


def get_application(app_name: str):
    return load_application(app_name)


def get_app_environment(app_name: str, env_name: str) -> Environment:
    application = get_application(app_name)
    application_environment = application.environments.get(env_name)

    if not application_environment:
        click.secho(
            f"The environment {env_name} was not found in the application {app_name}. "
            f"It either does not exist, or has not been deployed.",
            fg="red",
        )
        raise click.Abort

    return application_environment


def get_app_service(app_name: str, svc_name: str) -> Service:
    application = get_application(app_name)
    application_service = application.services.get(svc_name)

    if not application_service:
        click.secho(
            f"The service {svc_name} was not found in the application {app_name}. "
            f"It either does not exist, or has not been deployed.",
            fg="red",
        )
        raise click.Abort

    return application_service


def get_listener_rule_by_tag(elbv2_client, listener_arn, tag_key, tag_value):
    response = elbv2_client.describe_rules(ListenerArn=listener_arn)
    for rule in response["Rules"]:
        rule_arn = rule["RuleArn"]

        tags_response = elbv2_client.describe_tags(ResourceArns=[rule_arn])
        for tag_description in tags_response["TagDescriptions"]:
            for tag in tag_description["Tags"]:
                if tag["Key"] == tag_key and tag["Value"] == tag_value:
                    return rule


def get_vpc_id(session, env_name, vpc_name=None):
    if not vpc_name:
        vpc_name = f"{session.profile_name}-{env_name}"

    filters = [{"Name": "tag:Name", "Values": [vpc_name]}]
    vpcs = session.client("ec2").describe_vpcs(Filters=filters)["Vpcs"]

    if not vpcs:
        filters[0]["Values"] = [session.profile_name]
        vpcs = session.client("ec2").describe_vpcs(Filters=filters)["Vpcs"]

    if not vpcs:
        click.secho(
            f"No VPC found with name {vpc_name} in AWS account {session.profile_name}.", fg="red"
        )
        raise click.Abort

    return vpcs[0]["VpcId"]


def get_subnet_ids(session, vpc_id):
    subnets = session.client("ec2").describe_subnets(
        Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
    )["Subnets"]

    if not subnets:
        click.secho(f"No subnets found for VPC with id: {vpc_id}.", fg="red")
        raise click.Abort

    public_tag = {"Key": "subnet_type", "Value": "public"}
    public = [subnet["SubnetId"] for subnet in subnets if public_tag in subnet["Tags"]]
    private_tag = {"Key": "subnet_type", "Value": "private"}
    private = [subnet["SubnetId"] for subnet in subnets if private_tag in subnet["Tags"]]

    return public, private


def get_cert_arn(session, application, env_name):
    try:
        arn = find_https_certificate(session, application, env_name)
    except:
        click.secho(
            f"No certificate found with domain name matching environment {env_name}.", fg="red"
        )
        raise click.Abort

    return arn


def get_env_ips(vpc: str, application_environment: Environment) -> List[str]:
    account_name = f"{application_environment.session.profile_name}-vpc"
    vpc_name = vpc if vpc else account_name
    ssm_client = application_environment.session.client("ssm")

    try:
        param_value = ssm_client.get_parameter(Name=f"/{vpc_name}/EGRESS_IPS")["Parameter"]["Value"]
    except ssm_client.exceptions.ParameterNotFound:
        click.secho(f"No parameter found with name: /{vpc_name}/EGRESS_IPS")
        raise click.Abort

    return [ip.strip() for ip in param_value.split(",")]


@environment.command()
@click.option("--vpc-name", hidden=True)
@click.option("--name", "-n", required=True)
def generate(name, vpc_name):
    if vpc_name:
        click.secho(
            f"This option is deprecated. Please add the VPC name for your envs to {PLATFORM_CONFIG_FILE}",
            fg="red",
        )
        raise click.Abort

    try:
        conf = load_and_validate_platform_config()
    except SchemaError as ex:
        click.secho(f"Invalid `{PLATFORM_CONFIG_FILE}` file: {str(ex)}", fg="red")
        raise click.Abort

    env_config = apply_environment_defaults(conf)["environments"][name]
    profile_for_environment = env_config.get("accounts", {}).get("deploy", {}).get("name")
    click.secho(f"Using {profile_for_environment} for this AWS session")
    session = get_aws_session_or_abort(profile_for_environment)

    _generate_copilot_environment_manifests(name, conf["application"], env_config, session)


@environment.command(help="Generate terraform manifest for the specified environment.")
@click.option(
    "--name", "-n", required=True, help="The name of the environment to generate a manifest for."
)
@click.option(
    "--terraform-platform-modules-version",
    help=f"Override the default version of terraform-platform-modules. (Default version is '{DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION}').",
)
def generate_terraform(name, terraform_platform_modules_version):
    if not is_terraform_project():
        click.secho("This is not a terraform project. Exiting.", fg="red")
        exit(1)

    conf = load_and_validate_platform_config()

    env_config = apply_environment_defaults(conf)["environments"][name]
    _generate_terraform_environment_manifests(
        conf["application"], name, env_config, terraform_platform_modules_version
    )


def _generate_copilot_environment_manifests(name, application, env_config, session):
    env_template = setup_templates().get_template("env/manifest.yml")
    vpc_name = env_config.get("vpc", None)
    vpc_id = get_vpc_id(session, name, vpc_name)
    pub_subnet_ids, priv_subnet_ids = get_subnet_ids(session, vpc_id)
    cert_arn = get_cert_arn(session, application, name)
    contents = env_template.render(
        {
            "name": name,
            "vpc_id": vpc_id,
            "pub_subnet_ids": pub_subnet_ids,
            "priv_subnet_ids": priv_subnet_ids,
            "certificate_arn": cert_arn,
        }
    )
    click.echo(mkfile(".", f"copilot/environments/{name}/manifest.yml", contents, overwrite=True))


def _generate_terraform_environment_manifests(
    application, env, env_config, cli_terraform_platform_modules_version
):
    env_template = setup_templates().get_template("environments/main.tf")

    terraform_platform_modules_version = _determine_terraform_platform_modules_version(
        env_config, cli_terraform_platform_modules_version
    )

    contents = env_template.render(
        {
            "application": application,
            "environment": env,
            "config": env_config,
            "terraform_platform_modules_version": terraform_platform_modules_version,
        }
    )

    click.echo(mkfile(".", f"terraform/environments/{env}/main.tf", contents, overwrite=True))


def _determine_terraform_platform_modules_version(env_conf, cli_terraform_platform_modules_version):
    cli_terraform_platform_modules_version = cli_terraform_platform_modules_version
    env_conf_terraform_platform_modules_version = env_conf.get("versions", {}).get(
        "terraform-platform-modules"
    )
    version_preference_order = [
        cli_terraform_platform_modules_version,
        env_conf_terraform_platform_modules_version,
        DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION,
    ]
    return [version for version in version_preference_order if version][0]


def find_load_balancer(session: boto3.Session, app: str, env: str) -> str:
    lb_client = session.client("elbv2")

    describe_response = lb_client.describe_load_balancers()
    load_balancers = [lb["LoadBalancerArn"] for lb in describe_response["LoadBalancers"]]

    load_balancers = lb_client.describe_tags(ResourceArns=load_balancers)["TagDescriptions"]

    load_balancer_arn = None
    for lb in load_balancers:
        tags = {t["Key"]: t["Value"] for t in lb["Tags"]}
        if tags.get("copilot-application") == app and tags.get("copilot-environment") == env:
            load_balancer_arn = lb["ResourceArn"]

    if not load_balancer_arn:
        raise LoadBalancerNotFoundError()

    return load_balancer_arn


def find_https_listener(session: boto3.Session, app: str, env: str) -> str:
    load_balancer_arn = find_load_balancer(session, app, env)
    lb_client = session.client("elbv2")
    listeners = lb_client.describe_listeners(LoadBalancerArn=load_balancer_arn)["Listeners"]

    listener_arn = None

    try:
        listener_arn = next(l["ListenerArn"] for l in listeners if l["Protocol"] == "HTTPS")
    except StopIteration:
        pass

    if not listener_arn:
        raise ListenerNotFoundError()

    return listener_arn


def find_https_certificate(session: boto3.Session, app: str, env: str) -> str:
    listener_arn = find_https_listener(session, app, env)
    cert_client = session.client("elbv2")
    certificates = cert_client.describe_listener_certificates(ListenerArn=listener_arn)[
        "Certificates"
    ]

    certificate_arn = None

    try:
        certificate_arn = next(c["CertificateArn"] for c in certificates if c["IsDefault"])
    except StopIteration:
        raise CertificateNotFoundError()

    return certificate_arn


def find_target_group(app: str, env: str, svc: str, session: boto3.Session) -> str:
    rg_tagging_client = session.client("resourcegroupstaggingapi")
    response = rg_tagging_client.get_resources(
        TagFilters=[
            {
                "Key": "copilot-application",
                "Values": [
                    app,
                ],
                "Key": "copilot-environment",
                "Values": [
                    env,
                ],
                "Key": "copilot-service",
                "Values": [
                    svc,
                ],
            },
        ],
        ResourceTypeFilters=[
            "elasticloadbalancing:targetgroup",
        ],
    )
    for resource in response["ResourceTagMappingList"]:
        tags = {tag["Key"]: tag["Value"] for tag in resource["Tags"]}

        if (
            "copilot-service" in tags
            and tags["copilot-service"] == svc
            and "copilot-environment" in tags
            and tags["copilot-environment"] == env
            and "copilot-application" in tags
            and tags["copilot-application"] == app
        ):
            return resource["ResourceARN"]

    click.secho(
        f"No target group found for application: {app}, environment: {env}, service: {svc}",
        fg="red",
    )

    return None


def get_maintenance_page(session: boto3.Session, listener_arn: str) -> Union[str, None]:
    lb_client = session.client("elbv2")

    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]
    tag_descriptions = get_rules_tag_descriptions(rules, lb_client)

    maintenance_page_type = None
    for description in tag_descriptions:
        tags = {t["Key"]: t["Value"] for t in description["Tags"]}
        if tags.get("name") == "MaintenancePage":
            maintenance_page_type = tags.get("type")

    return maintenance_page_type


def delete_listener_rule(tag_descriptions: list, tag_name: str, lb_client: boto3.client):
    current_rule_arn = None

    for description in tag_descriptions:
        tags = {t["Key"]: t["Value"] for t in description["Tags"]}
        if tags.get("name") == tag_name:
            current_rule_arn = description["ResourceArn"]
            if current_rule_arn:
                lb_client.delete_rule(RuleArn=current_rule_arn)

    return current_rule_arn


def remove_maintenance_page(session: boto3.Session, listener_arn: str):
    lb_client = session.client("elbv2")

    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]
    tag_descriptions = get_rules_tag_descriptions(rules, lb_client)
    tag_descriptions = lb_client.describe_tags(ResourceArns=[r["RuleArn"] for r in rules])[
        "TagDescriptions"
    ]

    for name in ["MaintenancePage", "AllowedIps", "BypassIpFilter", "AllowedSourceIps"]:
        deleted = delete_listener_rule(tag_descriptions, name, lb_client)

        if name == "MaintenancePage" and not deleted:
            raise ListenerRuleNotFoundError()


def get_rules_tag_descriptions(rules: list, lb_client):
    tag_descriptions = []
    chunk_size = 20

    for i in range(0, len(rules), chunk_size):
        chunk = rules[i : i + chunk_size]
        resource_arns = [r["RuleArn"] for r in chunk]
        response = lb_client.describe_tags(ResourceArns=resource_arns)
        tag_descriptions.extend(response["TagDescriptions"])

    return tag_descriptions


def get_host_conditions(lb_client: boto3.client, listener_arn: str, target_group_arn: str):
    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]

    # Get current set of forwarding conditions for the target group
    for rule in rules:
        for action in rule["Actions"]:
            if action["Type"] == "forward" and action["TargetGroupArn"] == target_group_arn:
                conditions = rule["Conditions"]

    # filter to host-header conditions
    conditions = [
        {i: condition[i] for i in condition if i != "Values"}
        for condition in conditions
        if condition["Field"] == "host-header"
    ]

    # remove internal hosts
    conditions[0]["HostHeaderConfig"]["Values"] = [
        v for v in conditions[0]["HostHeaderConfig"]["Values"]
    ]

    return conditions


def create_header_rule(
    lb_client: boto3.client,
    listener_arn: str,
    target_group_arn: str,
    header_name: str,
    values: list,
    rule_name: str,
    priority: int,
):
    conditions = get_host_conditions(lb_client, listener_arn, target_group_arn)

    # add new condition to existing conditions
    combined_conditions = [
        {
            "Field": "http-header",
            "HttpHeaderConfig": {"HttpHeaderName": header_name, "Values": values},
        }
    ] + conditions

    lb_client.create_rule(
        ListenerArn=listener_arn,
        Priority=priority,
        Conditions=combined_conditions,
        Actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
        Tags=[
            {"Key": "name", "Value": rule_name},
        ],
    )

    click.secho(
        f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\n\nIf request header {header_name} contains one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.",
        fg="green",
    )


def create_source_ip_rule(
    lb_client: boto3.client,
    listener_arn: str,
    target_group_arn: str,
    values: list,
    rule_name: str,
    priority: int,
):
    conditions = get_host_conditions(lb_client, listener_arn, target_group_arn)

    # add new condition to existing conditions
    combined_conditions = [
        {
            "Field": "source-ip",
            "SourceIpConfig": {"Values": [value + "/32" for value in values]},
        }
    ] + conditions

    lb_client.create_rule(
        ListenerArn=listener_arn,
        Priority=priority,
        Conditions=combined_conditions,
        Actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
        Tags=[
            {"Key": "name", "Value": rule_name},
        ],
    )

    click.secho(
        f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\n\nIf request source ip matches one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.",
        fg="green",
    )


def add_maintenance_page(
    session: boto3.Session,
    listener_arn: str,
    app: str,
    env: str,
    services: List[Service],
    allowed_ips: tuple,
    template: str = "default",
):
    lb_client = session.client("elbv2")
    maintenance_page_content = get_maintenance_page_template(template)
    bypass_value = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))

    service_number = 1

    for svc in services:
        target_group_arn = find_target_group(app, env, svc.name, session)

        # not all of an application's services are guaranteed to have been deployed to an environment
        if not target_group_arn:
            continue

        allowed_ips = list(allowed_ips)
        max_allowed_ips = 100
        for ip_index, ip in enumerate(allowed_ips):
            forwarded_rule_priority = (service_number * max_allowed_ips) + ip_index
            create_header_rule(
                lb_client,
                listener_arn,
                target_group_arn,
                "X-Forwarded-For",
                [ip],
                "AllowedIps",
                forwarded_rule_priority,
            )
            create_source_ip_rule(
                lb_client,
                listener_arn,
                target_group_arn,
                [ip],
                "AllowedSourceIps",
                forwarded_rule_priority + 1,
            )

        bypass_rule_priority = service_number
        create_header_rule(
            lb_client,
            listener_arn,
            target_group_arn,
            "Bypass-Key",
            [bypass_value],
            "BypassIpFilter",
            bypass_rule_priority,
        )

        service_number += 1

        click.secho(
            f"\nUse a browser plugin to add `Bypass-Key` header with value {bypass_value} to your requests. For more detail, visit https://platform.readme.trade.gov.uk/activities/holding-and-maintenance-pages/",
            fg="green",
        )

    fixed_rule_priority = (service_number + 5) * max_allowed_ips
    lb_client.create_rule(
        ListenerArn=listener_arn,
        Priority=fixed_rule_priority,  # big number because we create multiple higher priority "AllowedIps" rules for each allowed ip for each service above.
        Conditions=[
            {
                "Field": "path-pattern",
                "PathPatternConfig": {"Values": ["/*"]},
            }
        ],
        Actions=[
            {
                "Type": "fixed-response",
                "FixedResponseConfig": {
                    "StatusCode": "503",
                    "ContentType": "text/html",
                    "MessageBody": maintenance_page_content,
                },
            }
        ],
        Tags=[
            {"Key": "name", "Value": "MaintenancePage"},
            {"Key": "type", "Value": template},
        ],
    )


def get_maintenance_page_template(template) -> str:
    template_contents = (
        Path(__file__)
        .parent.parent.joinpath(
            f"templates/svc/maintenance_pages/{template}.html",
        )
        .read_text()
        .replace("\n", "")
    )

    # [^\S]\s+ - Remove any space that is not preceded by a non-space character.
    return re.sub(r"[^\S]\s+", "", template_contents)


class CertificateNotFoundError(Exception):
    pass


class LoadBalancerNotFoundError(Exception):
    pass


class ListenerNotFoundError(Exception):
    pass


class ListenerRuleNotFoundError(Exception):
    pass
